from dash_extensions.enrich import FileSystemBackend
from kmviz.core.provider import Providers, make_provider_from_dict
from kmviz.core import KmVizError
from kmviz.core.plugin import installed_plugins, search_for_plugins
from kmviz.core.log import kmv_info
from kmviz.core.provider import PROVIDERS
from kmviz.core.metadata import METADBS
from importlib_resources import files

from kmviz.core.cache import make_server_cache_manager, make_callback_cache_manager, make_result_cache_manager

import shutil

class kState:
    def __init__(self):
        self._providers: Providers = Providers()
        self.dashboard_path = "/"
        self._manager = None
        self._cache = None
        self._plugins = {}
        self._external_css = []
        self._external_js = []
        self.plot_only = False
        self.session_only = False
        self._backend = None
        self._template = ""
        self._metatags = []
        self._limits = {
            "max_query": 2**32,
            "max_query_size": 2**32,
            "max_size": 2**64,
            "alphabet": "dna"
        }

        self._defaults = {
            "databases": [],
            "configuration": None,
            "hide_db": False,
            "hide_config": False
        }

        self._api = {
            "enabled": False,
            "route": "/api",
            "query_route": "/query",
            "limits": self._limits.copy()
        }

    def store_result(self, uid: str, results: tuple):
        self._cache.set(uid, results)

    def get_result(self, uid: str) -> tuple:
        res = self._cache.get(uid)
        if res is None:
            raise KmVizError("Not in cache.")
        return res

    def configure(self, config: dict):
        self._config = config
        self._configure_limits(config)
        self._configure_api(config)
        self._configure_template(config)
        self._configure_plugins(config)
        self._configure_providers(config)
        self._configure_defaults(config)
        self._configure_caches(config)

    @property
    def config(self):
        return self._config

    @property
    def plugins(self):
        return self._plugins

    @property
    def providers(self):
        return self._providers

    @property
    def manager(self):
        return self._manager

    @property
    def backend(self):
        return self._backend

    @property
    def css(self):
        return self._external_css

    @property
    def js(self):
        return self._external_js

    @property
    def template(self):
        return self._template

    @property
    def metatags(self):
        return self._metatags

    @property
    def limits(self):
        return self._limits

    @property
    def defaults(self):
        return self._defaults

    @property
    def api(self):
        return self._api["enabled"]

    @property
    def api_config(self):
        return self._api

    def instance_plugin(self):
        for name, p in self._plugins.items():
            if p.is_instance_plugin():
                if "dashboard" in p.config and p.config["dashboard"] != "/":
                    self.dashboard_path = p.config["dashboard"]
                else:
                    self.dashboard_path = "/dashboard"
                return p
        return None

    def _set_from_config(self, config, store):
        for k in store.keys():
            if k in config:
                if isinstance(store[k], dict):
                    self._set_from_config(config[k], store[k])
                else:
                    store[k] = config[k]

    def _configure_api(self, config: dict):
        if not "api" in config:
            return
        self._set_from_config(config["api"], self._api)

    def _configure_defaults(self, config: dict):
        if not "defaults" in config:
            return

        if "databases" in config["defaults"]:
            if isinstance(config["defaults"]["databases"], list):
                self._defaults["databases"] = config["defaults"]["databases"]
            else:
                self._defaults["databases"] = [config["defaults"]["databases"]]

            if "configuration" in config["defaults"]:
                self._defaults["configuration"] = config["defaults"]["configuration"]
            else:
                self._defaults["configuration"] = self._defaults["databases"][0]

            for d in self._defaults["databases"]:
                if d not in self.providers.list():
                    raise KmVizError(f"'{d}' database does not exist")

            dc = self._defaults["configuration"]
            if dc and dc not in self.providers.list():
                raise KmVizError(f"'{dc}' database does not exist")

            if "hide_db" in config["defaults"]:
                if config["defaults"]["hide_db"]:
                    if len(self._defaults["databases"]) != 1:
                        raise KmVizError("'hide_db=True' with multiple default databases")
                    self._defaults["hide_db"] = True

    def _configure_limits(self, config: dict):
        if not "input" in config:
            return
        self._set_from_config(config["input"], self._limits)
        #if "max_query" in config["input"]:
        #    self._limits["max_query"] = int(config["input"]["max_query"])
        #if "max_query_size" in config["input"]:
        #    self._limits["max_query_size"] = int(config["input"]["max_query_size"])
        #if "max_size" in config["input"]:
        #    self._limits["max_size"] = int(config["input"]["max_size"])
        #if "alphabet" in config["input"]:
        #    self._limits["alphabet"] = config["input"]["alphabet"]

    def _configure_template(self, config: dict):
        if "html" not in config:
            return

        if "template" in config["html"]:
            with open(config["html"]["template"], "r") as fin:
                self._template = fin.read()

        if "metatags" in config ["html"]:
            self._metatags = config["html"]["metatags"]
            if isinstance(self._metatags, dict):
                self._metatags = [self._metatags]

    def _configure_caches(self, config: dict):
        if "cache" not in config:
            config["cache"] = {}

        if "manager" not in config["cache"]:
            manager_default = { "type": "disk", "params": {"directory": "./kmviz_manager_cache"} }
            self._manager = make_callback_cache_manager(manager_default)
        else:
            self._manager = make_callback_cache_manager(config["cache"]["manager"])

        if "result" not in config["cache"]:
            result_params = {
                "cache_dir": "./kmviz_result_cache",
                "default_timeout": 60 * 60 * 24 * 14,
                "threshold": 0
            }
            result_default = { "type": "disk", "params": result_params }
            self._cache = make_result_cache_manager(result_default)
        else:
            self._cache = make_result_cache_manager(config["cache"]["result"])

        if "serverside" not in config["cache"]:
            server_params = {
                "cache_dir": "./kmviz_serverside_cache",
                "default_timeout": 60 * 60 * 24,
                "threshold": 0
            }
            server_default = { "type": "disk", "params": server_params }
            self._backend = make_server_cache_manager(server_default)
        else:
            self._backend = make_server_cache_manager(config["cache"]["serverside"])

    def _configure_plugins(self, config: dict):
        installed = installed_plugins()
        kmv_info(f"Installed plugins: {','.join(installed)}")

        if not "plugins" in config:
            self.dashboard_path = "/"
        else:
            self._plugins = search_for_plugins(config["plugins"])

            for name, plugin in self._plugins.items():
                if name in config["plugins"]:
                    plugin.configure(config["plugins"][name])

                self._external_css.extend(plugin.external_styles())
                self._external_js.extend(plugin.external_scripts())
                self._copy_plugin_assets(name)

                p_providers = plugin.providers()
                p_metadb = plugin.databases()

                if p_providers:
                    for pname, prov in p_providers:
                        if pname not in PROVIDERS:
                            kmv_info(f"Load external Provider '{pname}' from '{name}' plugin")
                            PROVIDERS[pname] = prov

                if p_metadb:
                    for pname, metadb in p_metadb:
                        if pname not in METADBS:
                            kmv_info(f"Load external MetaDB '{pname}' from '{name}' plugin")
                            METADBS[pname] = metadb

    def _copy_plugin_assets(self, plugin_name):
        p = files(plugin_name).joinpath("assets")
        main = str(files("kmviz").joinpath("assets"))
        if p.exists() and p.is_dir():
            shutil.copytree(p, f"{main}/_{plugin_name}_assets", dirs_exist_ok=True)

    def _configure_providers(self, config: dict):
        if "databases" not in config:
            raise KmVizError("'databases' section is missing in the configuration file.")

        for name, params in config["databases"].items():
            kmv_info(f"Load provider '{name}'")
            p = make_provider_from_dict(name, params)
            try:
                kmv_info(f"Init provider '{name}'")
                p.connect()
            except:
                raise KmVizError(f"Init fails for '{name}'")

            self._providers.add(p)

def init_state():
    global kmstate
    kmstate = kState()




from dash_extensions.enrich import html, dash_table, Input, Output, dcc, State, callback
import dash_mantine_components as dmc
from dash import no_update
import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
import plotly.io as pio

from dash_iconify import DashIconify

from kmviz.ui.utils import make_select_data, prevent_update_on_none, KMVIZ_ICONS
from kmviz.ui.components.figure import make_plot_title, make_plot_title_callbacks, make_select_input, make_hover_color_picker
from kmviz.ui.components.figure import make_axes, make_plot_legend, make_plot_shape
from kmviz.ui.components.figure import make_axes_callbacks, make_plot_legend_callbacks, make_plot_shape_callbacks
from kmviz.ui.components.select import kgsf
from kmviz.ui.components.store import ksf
from kmviz.ui.layouts.table import ktable

from kmviz.ui.id_factory import kmviz_factory as kf
import dash_bio as dashbio

import itertools
import operator
import random
import statistics

import warnings
warnings.filterwarnings('ignore', module='gradpyent')
from gradpyent import gradient
from gradpyent.library.formats import get_verified_color

kseq = kf.child("seq")

def blank_figure():
    fig = go.Figure(go.Scatter(x=[], y = []))
    fig.update_layout(template = "seaborn")
    fig.update_xaxes(showgrid = False, showticklabels = False, zeroline=False)
    fig.update_yaxes(showgrid = False, showticklabels = False, zeroline=False)

    return fig

def make_coverage(cov: list, abs, start, end):
    grad = gradient.Gradient(gradient_start=start, gradient_end=end)

    ranges = []
    series = []

    coverage = []
    legend = []

    for r in itertools.groupby(enumerate(cov), key=operator.itemgetter(1)):
        series.append(r[0])
        L = list(r[1])
        ranges.append((L[0][0], L[-1][0] + 1))

    colors = grad.get_gradient_series(series, fmt="html")

    for i, c in enumerate(colors):
        coverage.append({
            "bgcolor": c,
            "start": ranges[i][0],
            "end": ranges[i][1],
            "tooltip": round(series[i], 2),
            "onclick": round(series[i], 2),
        })

    if all(isinstance(a, int) for a in series):
        sorted_series = sorted(zip(series, colors))

        s = set()
        for value, color in sorted_series:
            if value not in s:
                legend.append({
                    "color": color,
                    "name": value
                })
                s.add(value)

    else:
        legend.append({
            "color": colors[min((val, idx) for (idx, val) in enumerate(series))[1]],
            "name": round(min(series), 2)
        })
        legend.append({
            "color": colors[max((val, idx) for (idx, val) in enumerate(series))[1]],
            "name": round(max(series), 2)
        })


    return coverage, legend

def make_sequence_layout():

    down_select = dmc.SegmentedControl(
        id=kseq.sid("down-select"),
        data=make_select_data(["png", "jpg", "svg", "pdf", "html", "json"]),
        value="png",
        size="xs",
        color=dmc.theme.DEFAULT_COLORS["dark"][0]
    )

    graph_view = html.Div([
        dmc.Space(h=5),
        dcc.Graph(
            figure=blank_figure(),
            id=kseq.sid("figure"),
            responsive=True,
            style = { "margin-left": "auto", "margin-right": "auto", "height": "50vh", "width": "95%" }
        ),

        dmc.Tabs([
            dmc.TabsList([
                dmc.Tab("Title", value="title"),
                dmc.Tab("Axes", value="axes"),
                dmc.Tab("Legend", value="legend"),
                dmc.Tab("Shape", value="shape"),
                dmc.Group([
                    down_select,
                    dmc.ActionIcon(
                        DashIconify(icon="material-symbols:download", width=20),
                        id=kseq.sid("down-button"),
                        variant="filled",
                        color = "#1C7ED6",
                    ),
                ], style = {"margin-left":"10px", "width": "fit-content"}, spacing=5),
            ]),
            dmc.TabsPanel(make_plot_title(kseq.child("title")), value="title"),
            dmc.TabsPanel(make_axes(kseq.child("axes")), value="axes"),
            dmc.TabsPanel(make_plot_legend(kseq.child("legend")), value="legend"),
            dmc.TabsPanel(make_plot_shape(kseq.child("shape")), value="shape"),
        ]),

        dcc.Download(id=kseq.sid("download")),
    ])

    seq_view = html.Div([
        dmc.Space(h=10),
        dmc.Group([
            dmc.SegmentedControl(
                id=kseq.sid("mode-select"),
                data=make_select_data(["base", "kmer"]),
                value="base",
                color=dmc.theme.DEFAULT_COLORS["dark"][0],
                size="xs"
            ),
            make_hover_color_picker(
                kseq("start-color"),
                "Min",
                icon=KMVIZ_ICONS["cback"],
                color=dict(hex="#E0DFDF")
            ),
            make_hover_color_picker(
                kseq("end-color"),
                "Max",
                icon=KMVIZ_ICONS["cback"],
                color=dict(hex="#FF0000")
            )
        ]),

        dmc.Space(h=10),

        html.Div([
            dashbio.SequenceViewer(
                id=kseq.sid("seq-view"),
                wrapAminoAcids=True,
                toolbar=False,
                search=False,
                badge=False,
                charsPerLine=130,
                sequenceMaxHeight="55vh",
            ),
        ], style = { "margin-left": "auto", "margin-right": "auto", "position":"relative" })

    ], style={"margin-left": "20px"})

    tabs = dmc.Tabs([
        dmc.TabsList([
            dmc.Tab("Graph view", value="graph", icon=DashIconify(icon="mdi:graph-line")),
            dmc.Tab("Sequence view", value="sequence", icon=DashIconify(icon="mdi:dna")),
            dmc.Group([
                dmc.Select(
                    id=kseq.sid("select"),
                    size="xs",
                    icon=DashIconify(icon="mingcute:document-fill"),
                    className="kmviz-gselect"
                ),
                dmc.ActionIcon(
                    DashIconify(icon="bi:filetype-json", width=20),
                    id=kseq.sid("down-data-button"),
                    variant="filled",
                    color = "#1C7ED6"
                ),
                dcc.Download(id=kseq.sid("download-data")),
            ]),
        ]),

        dmc.TabsPanel(graph_view, value="graph"),
        dmc.TabsPanel(seq_view, value="sequence"),
    ], value = "graph")

    return tabs

def to_bio_color(c):
    if not c:
        return c
    if "rgb" in c:
        m = c["rgb"]
        return [m["r"], m["g"], m["b"]]
    elif "hex" in c:
        return c["hex"]
    return c

def make_sequence_layout_callbacks():

    @callback(
        Input(kseq.sid("down-button"), "n_clicks"),
        State(kseq.sid("down-select"), "value"),
        State(kseq.sid("figure"), "figure"),
        Output(kseq.sid("download"), "data"),
        prevent_initial_call=True,
    )
    def download_cov(n_clicks, fmt, data):
        if n_clicks:
            if fmt == "html":
                content = pio.to_html(data, validate=False)
                return dict(content=content, filename="kmviz-cov.html")
            elif fmt == "json":
                content = pio.to_json(data, validate=False)
                return dict(content=content, filename="kmviz-cov.json")
            else:
                content = pio.to_image(data, fmt, validate=False)
                return dcc.send_bytes(content, f"kmviz-cov.{fmt}")

    @callback(
        Input(kseq.sid("down-data-button"), "n_clicks"),
        State(kgsf("provider"), "value"),
        State(kgsf("query"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("download-data"), "data"),
        prevent_initial_call=True,
    )
    def download_json(n_clicks, provider, query, query_result):
        if n_clicks:
            if provider.startswith("__kmviz_df"):
                prevent_update_on_none(None)

            prevent_update_on_none(provider, query)

            qr = query_result[query][provider]

            res = {}
            res[query] = {}

            for sample, resp in qr.response.items():
                if resp.has_abs():
                    res[query][sample] = {"ykmer": resp.covyk, "ybase": resp.covyb}
                else:
                    res[query][sample] = {"xkmer": resp.covxk, "xbase": resp.covxb}

            return dict(content=str(res), filename=f"{provider}-{query}.json")

    @callback(
        Input(kgsf("provider"), "value"),
        Input(kgsf("query"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("select"), "data"),
        Output(kseq.sid("select"), "value"),
        Output(kseq.sid("panel"), "disabled"),
        prevent_initial_call=True,
    )
    def update_sample_select(provider, query, query_result):
        prevent_update_on_none(provider, query)

        if provider.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        qr = query_result[query][provider]
        ids = list(qr.df["ID"])
        if not ids:
            return no_update, no_update, True
        return make_select_data(ids), ids[0], False

    @callback(
        Input(ktable.sid("grid"), "selectedRows"),
        Output(kseq.sid("select"), "value"),
        Output("tab-select", "value"),
        prevent_initial_call=True
    )
    def on_selected(data):
        if not len(data):
            prevent_update_on_none(None)

        return data[0]["ID"], "sequence"

    @callback(
        Input(kseq.sid("select"), "value"),
        Input(kgsf("provider"), "value"),
        Input(kgsf("query"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("figure"), "figure"),
        Output(kseq.sid("select"), "value"),
        prevent_initial_call=True,
    )
    def update_sequence_graph(sample, provider, query, query_result):
        if provider.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        prevent_update_on_none(sample, provider, query)

        qr = query_result[query][provider]

        if sample not in qr.response:
            sample = list(qr.df["ID"])[0]

        f = {'ID': {'filterType': 'text', 'type': 'contains', 'filter': sample}}

        index = list(range(len(qr.query.seq)))

        m = []
        cols = ["base"]

        if qr.response[sample].k:
            m = [None] * (qr.response[sample].k-1)
            cols = ["kmer", "base"]

        if qr.response[sample].has_abs():
            df = pd.DataFrame({"Sequence": index, "kmer": qr.response[sample].covyk + m, "base": qr.response[sample].covyb})
            fig = px.line(df, x="Sequence", y=cols, line_shape="linear", markers=True)
        else:
            df = pd.DataFrame({"Sequence": index, "kmer": qr.response[sample].covxk + m, "base": qr.response[sample].covxb})
            fig = px.line(df, x="Sequence", y=cols, line_shape="linear", markers=True)

        fig.update_layout(
            xaxis = dict(
                tickmode = 'array',
                tickvals = index,
                ticktext = [x for x in qr.query.seq]
            )
        )

        fig.update_layout(
            xaxis=dict(rangeslider=dict(visible=True,range=[0, len(qr.query.seq)]),
                       range=[0, 50],
                       type="category")
        )

        fig.update_layout(legend_title_text="")
        fig.update_xaxes(title_text="Sequence")
        fig.update_yaxes(rangemode="tozero")

        return fig, sample

    @callback(
        Input(kseq.sid("select"), "value"),
        Input(kgsf("provider"), "value"),
        Input(kgsf("query"), "value"),
        Input(kseq.sid("mode-select"), "value"),
        Input(kseq("start-color"), "value"),
        Input(kseq("end-color"), "value"),
        State(ksf("query-results"), "data"),
        Output(kseq.sid("seq-view"), "sequence"),
        Output(kseq.sid("seq-view"), "coverage"),
        Output(kseq.sid("seq-view"), "legend"),
        Output(kseq.sid("select"), "value"),
        prevent_initial_call=True,
    )
    def update_sequence_view(sample, provider, query, mode, start, end, query_result):
        if provider.startswith("__kmviz_df"):
            prevent_update_on_none(None)
        prevent_update_on_none(sample, provider, query)

        qr = query_result[query][provider]

        cov_type = "base"

        if sample not in qr.response:
            sample = list(qr.df["ID"])[0]

        if mode == "kmer" and not qr.response[sample].k:
            mode == "base"

        start = to_bio_color(start)
        end = to_bio_color(end)

        if qr.response[sample].has_abs():
            cov, leg = make_coverage(qr.response[sample].covyb if mode == "base" else qr.response[sample].covyk, True, start, end)
        else:
            cov, leg = make_coverage(qr.response[sample].covxb if mode == "base" else qr.response[sample].covxk, False, start, end)

        return qr.query.seq, cov, leg, sample

    make_plot_title_callbacks(kseq.child("title"), kseq.sid("figure"))
    make_axes_callbacks(kseq.child("axes"), kseq.sid("figure"))
    make_plot_legend_callbacks(kseq.child("legend"), kseq.sid("figure"))
    make_plot_shape_callbacks(kseq.child("shape"), kseq.sid("figure"))




import dash_mantine_components as dmc
from dash_extensions.enrich import Input, Output, State, html, callback, dcc
from dash import no_update
from kmviz.ui.utils import prevent_update_on_empty, make_select_data, prevent_update_on_none, icons
from kmviz.ui.id_factory import kmviz_factory as kf
from kmviz.ui.components.store import ksfr, ksf
from kmviz.ui import state
import orjson
from flask import jsonify
from dataclasses import dataclass
from kmviz.core.query import QueryResponseGeo
from typing import Dict

kpsf = kf.child("select")
kgsf = kf.child("global-select")

from dash_iconify import DashIconify

def make_select_provider():
    res = html.Div([
        dmc.Divider(size="sm", color="gray", label="DATABASES", labelPosition="center"),
        dmc.MultiSelect(
            id=kpsf("provider"),
            label="Database(s)",
            data=make_select_data(state.kmstate.providers.list()),
            clearable=True,
            withAsterisk=True,
            icon=DashIconify(icon="fluent:library-16-regular"),
            value=state.kmstate.defaults["databases"],
        )
    ], style = {"display": "none" if state.kmstate.defaults["hide_db"] else "inline"})
    return res

def make_select_provider_callbacks():
    @callback(
        Input(kpsf("provider"), "value"),
        Output(ksfr("provider-active"), "data"),
        Output(kf("select-provider-config"), "data"),
        Output(kf("select-provider-config"), "value"),
        prevent_initial_callbacks=True,
    )
    def select_provider(providers):
        data = None
        active = []
        default = None

        if providers:
            active = providers
            data = make_select_data(providers)
            default = active[0]

        return active, data, default

def make_select():
    return dmc.Group([
        dmc.ActionIcon(
            DashIconify(icon="bi:filetype-json", width=20),
            id=kf.sid("download-all-results-btn"),
            variant="filled",
            color = "#1C7ED6",
        ),
        dmc.Space(w=10),
        dcc.Download(id=kf.sid("download-all-results")),
        dmc.Switch(
            id=kf.sid("auto-apply"),
            offLabel=icons("autoff", width=20),
            onLabel=icons("auton", width=20),
            checked=True,
        ),
        dmc.Space(w=10),
        dmc.Select(
            id=kgsf("provider"),
            placeholder="Select database",
            icon=DashIconify(icon="fluent:library-16-regular"),
            className="kmviz-gselect",
            size="xs"
        ),
        dmc.Select(
            id=kgsf("query"),
            placeholder="Select query",
            icon=DashIconify(icon="mdi:dna"),
            className="kmviz-gselect",
            size="xs"
        ),
    ], spacing=3, align="center", className="kmviz-gselect-group")

def make_select_callbacks():

    @callback(
        Input(kgsf("provider"), "value"),
        Output(kf.child("map").sid("select-preset"), "data"),
        Output(kf.child("plot").sid("select-preset"), "data"),
        Output(kf.child("map").sid("select-preset"), "style"),
        Output(kf.child("plot").sid("select-preset"), "style"),
    )
    def update_map_preset(value):
        if not value:
            return None, None, {"display": "none"}, {"display": "none"}

        if value.startswith("__kmviz_df"):
            prevent_update_on_none(None)

        prevent_update_on_empty(state.kmstate.providers.list())

        p = state.kmstate.providers.get(value).presets

        map_data, map_style = None, {"display": "none"}
        plot_data, plot_style = None, {"display": "none"}
        if "map" in p:
            map_data = make_select_data(list(p["map"].keys()))
            map_style["display"] = "inline"
        if "plot" in p:
            plot_data = make_select_data(list(p["plot"].keys()))
            plot_style["display"] = "inline"

        return map_data, plot_data, map_style, plot_style

    @callback(
        Input(kf.sid("download-all-results-btn"), "n_clicks"),
        State(ksf("query-results"), "data"),
        Output(kf.sid("download-all-results"), "data")
    )
    def download_all(n_clicks, data):
        if n_clicks and len(data) and state.kmstate.providers.list():
            for qname, res in data.items():
                for name in res:
                    R = QueryResponseGeo(
                        res[name]._query,
                        res[name]._response,
                        orjson.loads(res[name].df.to_json()),
                        state.kmstate.providers.get(name).db.geodata
                    )
                    res[name] = R

            return dict(content=orjson.dumps(jsonify(data).json).decode(), filename=f"session.json")

        prevent_update_on_none(None)


from kmviz.ui import state
from kmviz.core.io import parse_fastx, KmVizIOError
from kmviz.core.query import Query, QueryResponseGeo

from flask import jsonify
from flask import request
import orjson
import uuid
import base64

import dataclasses

def api_root():
    res = dict(database={})

    for name, provider in state.kmstate.providers.all().items():
        options = list(provider.options.keys())
        res["database"][name] = {}

        options = {}
        for opt_name, opt in provider.options.items():
            options[opt_name] = {
                "type": type(opt.default).__name__,
                "state": dataclasses.asdict(opt)
            }
        res["database"][name]["options"] = options

    res["input"] = state.kmstate.api_config["limits"]

    return jsonify(res)

def make_query(fastx, on):
    uuid_str = f"kmviz-{str(uuid.uuid4())}"

    queries = [Query(name, seq) for name, seq, _ in parse_fastx(fastx, state.kmstate.api_config["limits"])]

    results = {}

    for provider in list(on.keys()):
        for k, v in state.kmstate.providers.get(provider).options.items():
            if k not in on[provider]:
                on[provider][k] = v.value
            else:
                on[provider][k] = type(v.value)(on[provider][k])

    for i, query in enumerate(queries):
        result = state.kmstate.providers.query(query, list(on.keys()), on, uuid_str)

        for name in result:
            R = QueryResponseGeo(
                result[name]._query,
                result[name]._response,
                orjson.loads(result[name].df.to_json()),
                state.kmstate.providers.get(name).db.geodata
            )
            result[name] = R

        results[query.name] = result

    return results

def api_query():
    try:
        if not "database" in request.form:
            return f"'database' entry is missing!", 400
        if not "fastx" in request.files:
            return f"'fastx' entry is missing!", 400

        options = {}
        for db in request.form.getlist("database"):
            if db not in state.kmstate.providers.list():
                return f"'{db}' is not a valid database", 400
            options[db] = {}
            for k in request.form:
                if k.startswith(db):
                    name = k.split("-", 1)[1]
                    if name not in state.kmstate.providers.get(db).options:
                        return f"'{name}' is not a valid option for '{db}'"
                    options[db][name] = request.form[k]


        res = make_query(request.files["fastx"].stream.read().decode(), options)
        return jsonify(res)
    except KmVizIOError as e:
        return f"Error: {str(e)}", 400
    except:
        return "An error occured while processing your request", 400


def register_api_routes(server):
    if state.kmstate.api:
        root = state.kmstate.api_config["route"]
        server.route(root, methods=["GET"])(api_root)
        server.route(root + state.kmstate.api_config["query_route"], methods=["POST"])(api_query)